/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>

#include <mpp/shmem.h>

#include <transpose_shmem.h>
#include <fftwnd_shmem.h>

/***************************** Plan Creation ****************************/

fftwnd_shmem_plan fftwnd_shmem_create_plan(int rank, const int *n,
					   fftw_direction dir,
					   int flags)
{
    fftwnd_shmem_plan p;

    if (rank < 2)
	return 0;

    p = (fftwnd_shmem_plan) fftw_malloc(sizeof(fftwnd_shmem_aux_data));
    p->p_fft = 0;
    p->p_transpose = 0;
    p->p_transpose_inv = 0;
    p->p_last = p->p_first = 0;

    p->p_fft = fftwnd_create_plan(rank, n, dir, flags | FFTW_IN_PLACE);
    if (!p->p_fft)
	fftwnd_shmem_destroy_plan(p);

    p->p_transpose = transpose_shmem_create_plan(n[0], n[1]);
    if (!p->p_transpose)
	fftwnd_shmem_destroy_plan(p);

    if (!p->p_transpose->output_in_place) {
	 p->p_last = fftw_create_plan(n[rank-1],dir,flags & (~FFTW_IN_PLACE));
	 if (!p->p_last)
	      fftwnd_shmem_destroy_plan(p);
    }

    p->p_transpose_inv = transpose_shmem_create_plan(n[1], n[0]);
    if (!p->p_transpose_inv)
	fftwnd_shmem_destroy_plan(p);

    if (!p->p_transpose_inv->output_in_place) {
	 p->p_first = fftw_create_plan(n[0],dir,flags & (~FFTW_IN_PLACE));
	 if (!p->p_first)
	      fftwnd_shmem_destroy_plan(p);
    }

    return p;
}

fftwnd_shmem_plan fftw2d_shmem_create_plan(int nx, int ny,
					   fftw_direction dir, int flags)
{
    int n[2];

    n[0] = nx;
    n[1] = ny;

    return fftwnd_shmem_create_plan(2, n, dir, flags);
}

fftwnd_shmem_plan fftw3d_shmem_create_plan(int nx, int ny, int nz,
					   fftw_direction dir, int flags)
{
    int n[3];

    n[0] = nx;
    n[1] = ny;
    n[2] = nz;

    return fftwnd_shmem_create_plan(3, n, dir, flags);
}

/********************** Plan Destruction ************************/

void fftwnd_shmem_destroy_plan(fftwnd_shmem_plan p)
{
     if (p) {
	  if (p->p_fft)
	       fftwnd_destroy_plan(p->p_fft);
	  if (p->p_transpose)
	       transpose_shmem_destroy_plan(p->p_transpose);
	  if (p->p_transpose_inv)
	       transpose_shmem_destroy_plan(p->p_transpose_inv);
	  if (p->p_first)
	       fftw_destroy_plan(p->p_first);
	  if (p->p_last)
	       fftw_destroy_plan(p->p_last);
	  free(p);
     }
}

/********************* Getting Local Size ***********************/

void fftwnd_shmem_local_sizes(fftwnd_shmem_plan p,
			      int *local_nx,
			      int *local_x_start,
			      int *local_ny_after_transpose,
			      int *local_y_start_after_transpose,
			      int *total_local_size)
{
    if (p) {
      transpose_shmem_get_local_size(p->p_transpose->nx,
				     p->p_transpose->my_pe,
				     p->p_transpose->n_pes,
				     local_nx,
				     local_x_start);
      transpose_shmem_get_local_size(p->p_transpose->ny,
				     p->p_transpose->my_pe,
				     p->p_transpose->n_pes,
				     local_ny_after_transpose,
				     local_y_start_after_transpose);
      *total_local_size =
	transpose_shmem_get_local_storage_size(p->p_transpose->nx,
					       p->p_transpose->ny,
					       p->p_transpose->my_pe,
					       p->p_transpose->n_pes);
      
      *total_local_size *= p->p_fft->n_after[1];
    }
}

/******************** Computing the Transform *******************/

void fftwnd_shmem(fftwnd_shmem_plan p, int n_fields, 
		  FFTW_COMPLEX *local_data, FFTW_COMPLEX *work,
		  fftwnd_shmem_output_order output_order)
{
    int fft_iter;
    int j, i;
    fftw_plan *plans;
    int *n, *n_before, *n_after;
    int nb, nx, ny, local_nx, local_ny;
    FFTW_COMPLEX *work_1d;
    int rank;

    if (n_fields <= 0)
	return;

    plans = p->p_fft->plans;
    n = p->p_fft->n;
    n_before = p->p_fft->n_before;
    n_after = p->p_fft->n_after;
    work_1d = p->p_fft->work;
    rank = p->p_fft->rank;
    nx = n[0];
    ny = n[1];
    local_nx = p->p_transpose->local_nx;
    local_ny = p->p_transpose->local_ny;

    for (fft_iter = 0; fft_iter < n_fields; ++fft_iter) {
	/* do other dimensions, except for first: */
	for (j = 1; j < rank - 1; ++j) {
	    nb = (n_before[j] / nx) * local_nx;
	    for (i = 0; i < nb; ++i)
		fftw(plans[j], n_after[j],
		local_data + fft_iter + i * n_fields * n[j] * n_after[j],
		     n_fields * n_after[j], n_fields,
		     work_1d, 1, 0);
	}

	/* do last dimension: */
	nb = (n_before[rank - 1] / nx) * local_nx;
	if (p->p_transpose->output_in_place)
	     fftw(plans[rank - 1], nb,
		  local_data + fft_iter, n_fields, n[rank - 1] * n_fields,
		  work_1d, 1, 0);
	else
	     fftw(p->p_last, nb,
		  local_data + fft_iter, n_fields, n[rank - 1] * n_fields,
		  work + fft_iter, n_fields, n[rank - 1] * n_fields);
    }

    /* transpose the first two indices: */
    if (p->p_transpose->output_in_place)
	 transpose_shmem(p->p_transpose,
			 (FFTW_REAL *) local_data, 
			 (FFTW_REAL *) work,
			 2 * n_fields * n_after[1]);
    else
	 transpose_shmem(p->p_transpose,
			 (FFTW_REAL *) work,
			 (FFTW_REAL *) local_data, 
			 2 * n_fields * n_after[1]);

    for (fft_iter = 0; fft_iter < n_fields; ++fft_iter) {
	/* do first dimension: */
	nb = local_ny;
	if (p->p_transpose_inv->output_in_place ||
	    output_order != FFTW_NORMAL_ORDER)
	     for (i = 0; i < nb; ++i)
		  fftw(plans[0], n_after[1],
		       local_data + fft_iter + i * n_fields * n[0]*n_after[1],
		       n_fields * n_after[1], n_fields,
		       work_1d, 1, 0);
	else
	     for (i = 0; i < nb; ++i)
		  fftw(p->p_first, n_after[1],
		       local_data + fft_iter + i * n_fields * n[0]*n_after[1],
		       n_fields * n_after[1], n_fields,
		       work + fft_iter + i * n_fields * n[0]*n_after[1],
		       n_fields * n_after[1], n_fields);
    }

    /* transpose back, if desired: */
    if (output_order == FFTW_NORMAL_ORDER) {
	if (p->p_transpose_inv->output_in_place)
	     transpose_shmem(p->p_transpose_inv, 
			     (FFTW_REAL *) local_data,
			     (FFTW_REAL *) work,
			     2 * n_fields * n_after[1]);
	else
	     transpose_shmem(p->p_transpose_inv, 
			     (FFTW_REAL *) work,
			     (FFTW_REAL *) local_data,
			     2 * n_fields * n_after[1]);
   }
}
