/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *
 */

#include <stdio.h>
#include <math.h>

#include <mpp/shmem.h>
#include <mpp/sync_proto.h>

#include <fftwnd_shmem.h>

#define NUM_ITER 90000000L

#define N_TESTS_3D 8

extern void initialize_fft_data(FFTW_COMPLEX * arr, long n);

int main(int argc, char **argv)
{
    int n3[N_TESTS_3D][3] = 
    {
	 { 16, 16, 16 },
	 { 24, 24, 24 },
	 { 32, 32, 32 },
	 { 48, 48, 48 },
	 { 64, 64, 64 },
	 { 80, 80, 80 },
	 {100,100,100 },
	 { 128, 128, 128 },
/*         {256, 256, 256 }, */
    };
    int i, test, iter, max_iter;
    fftw_time start_t, end_t, init_t;
    FFTW_COMPLEX *cin, *cin_shmem, *work;
    double time_scale,time1,time2;
    int max_size;
    fftwnd_plan plan_nd;
    fftwnd_shmem_plan plan_nd_shmem;
    int my_pe;

    my_pe = shmem_my_pe();

    /*************** Benchmark fftwnd_shmem ****************/
     if (my_pe == 0)
    printf("\n");

    max_size = 0;
    for (i = 0; i < N_TESTS_3D; ++i)
	if (n3[i][0]*n3[i][1]*n3[i][2] > max_size)
	    max_size = n3[i][0]*n3[i][1]*n3[i][2];

    if (my_pe == 0) {
	 cin = malloc(max_size * sizeof(FFTW_COMPLEX));
	 
	 if (!cin) {
	      printf("Not enough memory!  At least %d bytes needed.\n",
		     max_size * sizeof(FFTW_COMPLEX));
	      exit(1);
	 }
    }
    else
      cin = 0;

     if (my_pe == 0)
       printf("%11s%12s%13s%13s%17s%13s\n", 
	      "Array Size", "FFTWND", "FFTWND_SHMEM",
	      "Speedup","FFTWND_SHMEM (T)","Speedup (T)");

    for (test = 0; test < N_TESTS_3D; ++test) {
	int N;
	int local_nx, local_x_start, local_ny, local_y_start, local_N;

	if (my_pe == 0)
	     plan_nd = fftwnd_create_plan(3,n3[test], FFTW_FORWARD, 
					  FFTW_IN_PLACE | FFTW_MEASURE);
	plan_nd_shmem = fftwnd_shmem_create_plan(3,n3[test], FFTW_FORWARD, 
					     FFTW_IN_PLACE | FFTW_MEASURE);

	fftwnd_shmem_local_sizes(plan_nd_shmem,&local_nx,&local_x_start,
				 &local_ny,&local_y_start,&local_N);

	cin_shmem = (FFTW_COMPLEX *) shmalloc(local_N * sizeof(FFTW_COMPLEX));
	work = (FFTW_COMPLEX *) malloc(local_N * sizeof(FFTW_COMPLEX));

	if (!cin_shmem || !work) {
	  printf("Out of memory on processor %d!n",my_pe);
	  exit(1);
	}

	N = n3[test][0]*n3[test][1]*n3[test][2];
	local_N = local_nx * n3[test][1]*n3[test][2];

	max_iter = NUM_ITER / (N * log(2.0 * N));

	if (max_iter < 1)
	     max_iter = 1;

	time_scale = 1.0e6 / (max_iter * (log(N)/log(2.0) * N));

	if (my_pe == 0) {
	     initialize_fft_data(cin, N);
	     start_t = fftw_get_time();
	     for (iter = 0; iter < max_iter; ++iter)
		  initialize_fft_data(cin, N);
	     end_t = fftw_get_time();
	     init_t = end_t - start_t;
	     
	     {
		  char s[20];
		  sprintf(s,"%dx%dx%d",n3[test][0],n3[test][1],n3[test][2]);
		  if (my_pe == 0)
		       printf("%11s",s);
	     }
	     fflush(stdout);
	     
	     /* Time FFTW: */
	     
	     initialize_fft_data(cin, N);
	     fftwnd(plan_nd, 1, cin, 1, 0, 0, 0, 0);
	     start_t = fftw_get_time();
	     for (iter = 0; iter < max_iter; ++iter) {
		  initialize_fft_data(cin, N);
		  fftwnd(plan_nd, 1, cin, 1, 0, 0, 0, 0);
	     }
	     end_t = fftw_get_time();
	     if (my_pe == 0) {
		  printf("%12g", 
			 time1=fftw_time_to_sec(end_t - start_t - init_t) *
			 time_scale);
		  fflush(stdout);
	     }
	}
	else
	     time1 = 0;

	/* Time SHMEM FFTW: */

	initialize_fft_data(cin_shmem, local_N);
	start_t = fftw_get_time();
	for (iter = 0; iter < max_iter; ++iter)
	    initialize_fft_data(cin_shmem, local_N);
	end_t = fftw_get_time();
	init_t = end_t - start_t;

	initialize_fft_data(cin_shmem, local_N);
	fftwnd_shmem(plan_nd_shmem, 1, cin_shmem, work, FFTW_NORMAL_ORDER);
	barrier();
	start_t = fftw_get_time();
	for (iter = 0; iter < max_iter; ++iter) {
	    initialize_fft_data(cin_shmem, local_N);
	    fftwnd_shmem(plan_nd_shmem, 1, cin_shmem, work, FFTW_NORMAL_ORDER);
	}
	end_t = fftw_get_time();
	if (my_pe == 0) {
	printf("%13g", time2=fftw_time_to_sec(end_t - start_t - init_t) *
                       time_scale);
	printf("%13g",time1/time2);
	fflush(stdout);
	}

	/* Time SHMEM FFTW (transposed output): */

	initialize_fft_data(cin_shmem, local_N);
	fftwnd_shmem(plan_nd_shmem, 1, cin_shmem, work, FFTW_TRANSPOSED_ORDER);
	barrier();
	start_t = fftw_get_time();
	for (iter = 0; iter < max_iter; ++iter) {
	    initialize_fft_data(cin_shmem, local_N);
	    fftwnd_shmem(plan_nd_shmem, 1, cin_shmem, work, 
			 FFTW_TRANSPOSED_ORDER);
	}
	end_t = fftw_get_time();
	if (my_pe == 0) {
	printf("%17g", time2=fftw_time_to_sec(end_t - start_t - init_t) *
                       time_scale);
	printf("%13g\n",time1/time2);
	fflush(stdout);
	}

	/* Done. */

	if (my_pe == 0)
	     fftwnd_destroy_plan(plan_nd);
	fftwnd_shmem_destroy_plan(plan_nd_shmem);

	shfree(cin_shmem);
	free(work);
    }

    if (my_pe == 0)
	 free(cin);

    return 0;
}

void initialize_fft_data(FFTW_COMPLEX * arr, long n)
{
    long i;

    for (i = 0; i < n; i++) { /* initialize to some arbitrary values: */
	c_re(arr[i]) = 0.56923456;
	c_im(arr[i]) = 0.23858572;
    }
}
