/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>
#include <stdio.h>
#include <mpp/shmem.h>
#include <mpp/sync_proto.h>

#include <transpose_shmem.h>

#define NX 14
#define NY 10

int main(int argc, char **argv)
{
     int my_pe,n_pes;
     int i,j;
     int nx,ny;
     int local_x_start,local_nx,local_y_start,local_ny, xbsize,ybsize;
     int total_local_size;
     double *work, *data, *input, *output, *param2;
     transpose_shmem_plan tp;
     int el_size = 1, el;

     my_pe = shmem_my_pe();
     n_pes = shmem_n_pes();

     if (argc < 3) {
	  if (my_pe == 1)
	       printf("Syntax: test_transpose_shmem nx ny [el_size]\n");
	  return 1;
     }

     nx = atoi(argv[1]);
     ny = atoi(argv[2]);

     if (argc > 3)
	  el_size = atoi(argv[3]);

     transpose_shmem_get_local_size(nx,my_pe,n_pes,
				  &local_nx,&local_x_start);
     transpose_shmem_get_local_size(ny,my_pe,n_pes,
				  &local_ny,&local_y_start);
     total_local_size = transpose_shmem_get_local_storage_size(nx,ny,
							     my_pe,n_pes);


     if (total_local_size == 0) {
       work = 0;
       data = 0;
     }
     else {
       /* allocate twice the required storage to check for overwrites: */
       work = (double*)malloc(2*total_local_size*el_size*sizeof(double));
       data = (double*)shmalloc(2*total_local_size*el_size*sizeof(double));

       if (total_local_size > 0 && (!work || !data)) {
	 printf("Out of memory on processor %d!\n",my_pe);
	 exit(1);
       }
     }

     tp = transpose_shmem_create_plan(nx,ny);
     
     if (tp->output_in_place) {
	  input = data;
	  output = data;
	  param2 = work;
     }
     else {
	  input = work;
	  output = data;
	  param2 = output;
     }

     for (i = 0; i < 2*total_local_size*el_size; ++i)
       work[i] = data[i] = -1.0; /* initialize to dummy value */

     for (i = 0; i < local_nx; ++i)
	  for (j = 0; j < ny; ++j)
	       for (el = 0; el < el_size; ++el)
		    input[el+el_size*(i*ny+j)] = ny * (i + local_x_start) + j;

     /* output plan data, one process at a time: */
     for (j = 0; j < n_pes; ++j) {
	  if (j == my_pe) {
	       printf("Plan for process %d:\n",j);
	       printf("   nx = %d, ny = %d\n",tp->nx,tp->ny);
	       printf("   local_nx = %d, local_ny = %d\n",
		      tp->local_nx,tp->local_ny);

	       if (local_nx > 0 || local_ny > 0) {
	       printf("   num_blocks = %d\n",
		      tp->num_blocks);
	       for (i = 0; i < tp->num_blocks; ++i)
		    printf("      block_dest[%d]: dest_pe = %d, "
			   "num_cols = %d, src_offset = %d\n",
			   i,
			   tp->block_dest[i].dest_pe,
			   tp->block_dest[i].num_cols,
			   tp->block_dest[i].src_offset);
	       }

	       printf("\n");
	  }
	  barrier();
     }
	      
     transpose_shmem(tp,input,param2,el_size);

     if (my_pe == 0)
	  printf("\nComputing transpose of %dx%d matrix (%d elements)...\n",
		 nx,ny,el_size);

     barrier();
     
     printf("Checking result on process %d...\n",my_pe);

     for (i = 0; i < local_ny; ++i)
	  for (j = 0; j < nx; ++j) 
	       for (el = 0; el < el_size; ++el) {
		    if (output[el+el_size*(i*nx + j)] != 
			(double) (j*ny + i+local_y_start)) {
			 printf("Error with x=%d, y=%d on process %d!\n"
				"  -- is %g rather than %g\n",
				j,i+local_y_start,my_pe,
				output[el+el_size*(i*nx + j)], 
				(double) (j*ny + i+local_y_start));
			 exit(1);
	       }
	  }

     /* Check for writes past the end of the data arrays: */
     for (i = total_local_size*el_size; i < 2*total_local_size*el_size; ++i) {
       if (work[i] != -1.0) {
	 printf("Processor %d: work data overwritten at %d with %g!\n",
		my_pe,i,work[i]);
	 exit(1);
       }
       if (data[i] != -1.0) {
	 printf("Processor %d: input data overwritten at %d with %g!\n",
		my_pe,i,data[i]);
	 exit(1);
       }
     }

     if (work)
	  free(work);
     if (data)
	  shfree(data);

     printf("Process %d okay!\n",my_pe);
     
     return 0;
}

