    /*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>
#include <string.h>

#include <mpp/shmem.h>
#include <mpp/sync_proto.h>

#include <transpose_shmem.h>

static int transpose_shmem_get_block_size(int n, int n_pes)
{
     if (n_pes > n)
	  n_pes = n;

     return((n + n_pes - 1) / n_pes);
}

void transpose_shmem_get_local_size(int n, int my_pe, int n_pes,
				  int *local_n, int *local_start)
{
     int block_size;

     if (n_pes > n)
	  n_pes = n;

     block_size = transpose_shmem_get_block_size(n,n_pes);
     *local_start = my_pe * block_size;

     if (*local_start >= n)
	  *local_n = 0;
     else if (*local_start + block_size >= n)
	  *local_n = n - *local_start;
     else
	  *local_n = block_size;
}

int transpose_shmem_get_local_storage_size(int nx, int ny,
					 int my_pe, int n_pes)
{
     int x_block_size,y_block_size;

     /* Note: unlike for MPI, the allocated data on every
	processor MUST be the same size to use shmem properly.  */

     x_block_size = transpose_shmem_get_block_size(nx,n_pes);
     y_block_size = transpose_shmem_get_block_size(ny,n_pes);

     if (x_block_size*ny > nx*y_block_size)
	  return (x_block_size*ny);
     else
	  return (nx*y_block_size);
}

static int sort_block_compare(int src_pe, int dest_pe1, int dest_pe2)
{
     int xor1,xor2;

     xor1 = src_pe ^ dest_pe1;
     xor2 = src_pe ^ dest_pe2;
     if (xor1 > xor2)
	  return 1;
     else if (xor1 < xor2)
	  return -1;
     else
	  return 0;
}

static void sort_block_dests(int my_pe,
			     transpose_shmem_block_dest *block_dest,
			     int num_blocks)
{
     transpose_shmem_block_dest swap;
     int i,j;

     /* Do a simple N^2 sort of the blocks: */
     for (i = 0; i < num_blocks-1; ++i)
	  for (j = i+1; j < num_blocks; ++j)
	       if (sort_block_compare(my_pe,
				      block_dest[i].dest_pe,
				      block_dest[j].dest_pe) > 0) {
		    swap = block_dest[i];
		    block_dest[i] = block_dest[j];
		    block_dest[j] = swap;
	       }
}

#define WHICH_PE(dest_i,i_block_size) ((dest_i)/(i_block_size))

transpose_shmem_plan transpose_shmem_create_plan(int nx, int ny)
{
     transpose_shmem_plan p;
     int my_pe, n_pes;
     int x_block_size, local_nx, local_x_start;
     int y_block_size, local_ny, local_y_start;
     transpose_shmem_block_dest *block_dest = 0;
     int block, block_size = 0, num_blocks = 0;
     int block_size_transpose = 0;
     int *perm_block_dest = 0;
     int num_perm_blocks = 0, perm_block_size = 0, perm_block;
     int move_size = 0;
     char *move = 0;

     my_pe = shmem_my_pe();
     n_pes = shmem_n_pes();

     x_block_size = transpose_shmem_get_block_size(nx,n_pes);
     transpose_shmem_get_local_size(nx,my_pe,n_pes,&local_nx,&local_x_start);
     y_block_size = transpose_shmem_get_block_size(ny,n_pes);
     transpose_shmem_get_local_size(ny,my_pe,n_pes,&local_ny,&local_y_start);

     if (local_nx > 0 || local_ny > 0) {

     num_blocks = n_pes;

     block_dest = malloc(num_blocks * sizeof(transpose_shmem_block_dest));
     if (!block_dest) {
	  return 0;
     }

     for (block = 0; block < num_blocks; ++block) {
	  block_dest[block].dest_pe = block;

	  if (block_dest[block].dest_pe == my_pe) {
	       block_dest[block].src_offset = local_y_start;
	       block_dest[block].num_cols = local_ny;
	  }
	  else {
	       int dest_ny, dest_y_start;

	       transpose_shmem_get_local_size(ny,block_dest[block].dest_pe,
					      n_pes,&dest_ny,&dest_y_start);
	       block_dest[block].src_offset = dest_y_start;
	       block_dest[block].num_cols = dest_ny;
	  }
     }

     sort_block_dests(my_pe,block_dest,num_blocks);

     }

     p = malloc(sizeof(transpose_shmem_plan_struct));
     if (!p) {
          free(block_dest);
	  return 0;
     }

     p->output_in_place = 0;

     p->nx = nx; p->ny = ny;
     p->local_nx = local_nx; p->local_ny = local_ny;
     p->local_x_start = local_x_start;

     p->my_pe = my_pe; p->n_pes = n_pes;

     p->block_dest = block_dest;
     p->num_blocks = num_blocks;

     return p;
}

void transpose_shmem_destroy_plan(transpose_shmem_plan p)
{
     if (p) {
	  if (p->block_dest)
	       free(p->block_dest);
	  free(p);
     }
}

void transpose_shmem(transpose_shmem_plan p, 
		     double *local_data, double *local_data_t,
		     int el_size)
{
     int my_pe, n_pes;
     int nx, ny, local_nx, local_ny, local_x_start;
     transpose_shmem_block_dest *block_dest;
     int block_size, num_blocks, block;
     int *perm_block_dest, num_perm_blocks, perm_block_size;

     my_pe = p->my_pe;
     n_pes = p->n_pes;
     nx = p->nx;
     ny = p->ny;
     local_nx = p->local_nx;
     local_ny = p->local_ny;
     local_x_start = p->local_x_start;

     block_dest = p->block_dest;
     num_blocks = p->num_blocks;

     shmem_set_cache_inv(); /* nop except on T3D */

     barrier();

     if (local_nx > 0)
	  for (block = 0; block < num_blocks; ++block) 
	       if (block_dest[block].num_cols > 0) {
		    int row, col, ncols, src_offset;
		    
		    ncols = block_dest[block].num_cols;
		    src_offset = block_dest[block].src_offset;
		    
		    if (block_dest[block].dest_pe == my_pe)
			 for (row = 0; row < local_nx; ++row)
			      for (col = 0; col < ncols; ++col)
				   memcpy(local_data_t +
					  el_size * (local_x_start +
						     row + col * nx),
					  local_data +
					  el_size * (src_offset +
						     row * ny + col),
					  el_size * sizeof(double));
		    else
			 for (row = 0; row < local_nx; ++row)
			      for (col = 0; col < ncols; ++col)
				   shmem_put(local_data_t +
					     el_size * (local_x_start +
							row + col * nx),
					     local_data +
					     el_size * (src_offset +
							row * ny + col),
					     el_size,
					     block_dest[block].dest_pe);
	       }
     
     barrier(); /* wait for completion */

     shmem_clear_cache_inv(); /* a nop except on the T3D */
}

