/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>
#include <string.h>

#include <mpp/shmem.h>
#include <mpp/sync_proto.h>

#include <transpose_shmem.h>

#include "TOMS_transpose.h"

static int transpose_shmem_get_block_size(int n, int n_pes)
{
     if (n_pes > n)
	  n_pes = n;

     return((n + n_pes - 1) / n_pes);
}

static int gcd(int a, int b)
{
	int r;
	do {
		r = a % b;
		a = b;
		b = r;
	} while (r != 0);
	return a;
}

void transpose_shmem_get_local_size(int n, int my_pe, int n_pes,
				  int *local_n, int *local_start)
{
     int block_size;

     if (n_pes > n)
	  n_pes = n;

     block_size = transpose_shmem_get_block_size(n,n_pes);
     *local_start = my_pe * block_size;

     if (*local_start >= n)
	  *local_n = 0;
     else if (*local_start + block_size >= n)
	  *local_n = n - *local_start;
     else
	  *local_n = block_size;
}

int transpose_shmem_get_local_storage_size(int nx, int ny,
					 int my_pe, int n_pes)
{
     int x_block_size,y_block_size;

     /* Note: unlike for MPI, the allocated data on every
	processor MUST be the same size to use shmem properly.  */

     x_block_size = transpose_shmem_get_block_size(nx,n_pes);
     y_block_size = transpose_shmem_get_block_size(ny,n_pes);

     if (x_block_size*ny > nx*y_block_size)
	  return (x_block_size*ny);
     else
	  return (nx*y_block_size);
}

static int sort_block_compare(int src_pe, int dest_pe1, int dest_pe2)
{
     int xor1,xor2;

     xor1 = src_pe ^ dest_pe1;
     xor2 = src_pe ^ dest_pe2;
     if (xor1 > xor2)
	  return 1;
     else if (xor1 < xor2)
	  return -1;
     else
	  return 0;
}

static void sort_block_dests(int my_pe,
			     transpose_shmem_block_dest *block_dest,
			     int num_blocks)
{
     transpose_shmem_block_dest swap;
     int i,j;

     /* Do a simple N^2 sort of the blocks: */
     for (i = 0; i < num_blocks-1; ++i)
	  for (j = i+1; j < num_blocks; ++j)
	       if (sort_block_compare(my_pe,
				      block_dest[i].dest_pe,
				      block_dest[j].dest_pe) > 0) {
		    swap = block_dest[i];
		    block_dest[i] = block_dest[j];
		    block_dest[j] = swap;
	       }
}

#define WHICH_PE(dest_i,i_block_size) ((dest_i)/(i_block_size))

transpose_shmem_plan transpose_shmem_create_plan(int nx, int ny)
{
     transpose_shmem_plan p;
     int my_pe, n_pes;
     int x_block_size, local_nx, local_x_start;
     int y_block_size, local_ny, local_y_start;
     transpose_shmem_block_dest *block_dest = 0;
     int block, block_size = 0, num_blocks = 0;
     int block_size_transpose = 0;
     int *perm_block_dest = 0;
     int num_perm_blocks = 0, perm_block_size = 0, perm_block;
     int move_size = 0;
     char *move = 0;

     my_pe = shmem_my_pe();
     n_pes = shmem_n_pes();

     x_block_size = transpose_shmem_get_block_size(nx,n_pes);
     transpose_shmem_get_local_size(nx,my_pe,n_pes,&local_nx,&local_x_start);
     y_block_size = transpose_shmem_get_block_size(ny,n_pes);
     transpose_shmem_get_local_size(ny,my_pe,n_pes,&local_ny,&local_y_start);

     if (local_nx > 0 || local_ny > 0) {

     perm_block_size = gcd(nx,x_block_size);
     num_perm_blocks = (nx / perm_block_size) * local_ny;
     {
	  int tmp = (local_nx / perm_block_size) * ny;
	  if (tmp > num_perm_blocks)
	       num_perm_blocks = tmp;
     }

     perm_block_dest = malloc(sizeof(int) * num_perm_blocks);
     if (!perm_block_dest)
	  return 0;

     for (perm_block = 0; perm_block < num_perm_blocks; ++perm_block)
	  perm_block_dest[perm_block] = num_perm_blocks; 

     block_size = local_nx * y_block_size;

     num_blocks = n_pes;

     block_dest = malloc(num_blocks * sizeof(transpose_shmem_block_dest));
     if (!block_dest) {
	  free(perm_block_dest);
	  return 0;
     }

     block_size_transpose = local_ny * x_block_size;

     for (block = 0; block < num_blocks; ++block) {
	  int block_x, block_y, block_new_x, block_new_y;
	  int num_perm_blocks_received, num_perm_rows_received;
	  int dest_pe;
	  int recv_size;
	  
	  block_x = local_x_start;
	  block_y = block * y_block_size;

	  block_dest[block].block_num = block;

	  if (block_y >= ny)
	       block_dest[block].send_size = 0;
	  else if (block_y + y_block_size >= ny) /* the last block */
	       block_dest[block].send_size = local_nx*ny - block*block_size;
	  else
	       block_dest[block].send_size = block_size;

	  dest_pe = block;

	  if (dest_pe == my_pe) {
	       block_dest[block].dest_pe = dest_pe;
	       recv_size = 
		    block_dest[block].send_size;
	       block_new_x = block_x;
	       block_new_y = block_y;
	       block_dest[block].dest_offset = block_size_transpose * block;
	  }
	  else {
	       int dest_nx, dest_x_start;

	       transpose_shmem_get_local_size(nx,dest_pe,n_pes,
					    &dest_nx,&dest_x_start);
	       block_dest[block].dest_pe = dest_pe;
	       block_new_x = dest_x_start;
	       block_new_y = local_y_start;
	       recv_size = dest_nx * local_ny;
	       
	       transpose_shmem_get_local_size(ny,dest_pe,n_pes,
					    &dest_nx,&dest_x_start);
	       block_dest[block].dest_offset = dest_nx * x_block_size *
		    (local_x_start / x_block_size);
	  }
	  
	  if (recv_size > 0) {
	       num_perm_blocks_received = recv_size /
		    perm_block_size;
	       num_perm_rows_received = num_perm_blocks_received / local_ny;

	       for (perm_block = 0; perm_block < num_perm_blocks_received;
		    ++perm_block) {
		    int old_block, new_block;
		    
		    old_block = perm_block + 
			 block * (block_size_transpose / perm_block_size);
		    new_block = perm_block % num_perm_rows_received +
			 block_new_x / perm_block_size +
			 (block_new_y - local_y_start + 
			  perm_block / num_perm_rows_received) 
			 * (nx / perm_block_size);
		    
		    perm_block_dest[old_block] = new_block;
	       }
	  }
     }

     sort_block_dests(my_pe,block_dest,num_blocks);

     move_size = (local_nx + ny)/ 2;
     move = malloc(move_size*sizeof(char));
     if (!move) {
          free(block_dest);
          free(perm_block_dest);
          return 0;
     }

     }

     p = malloc(sizeof(transpose_shmem_plan_struct));
     if (!p) {
          free(block_dest);
	  free(perm_block_dest);
	  free(move);
	  return 0;
     }

     p->output_in_place = 1;

     p->nx = nx; p->ny = ny;
     p->local_nx = local_nx; p->local_ny = local_ny;

     p->my_pe = my_pe; p->n_pes = n_pes;

     p->block_dest = block_dest;
     p->block_size = block_size;
     p->block_size_transpose = block_size_transpose;
     p->num_blocks = num_blocks;

     p->perm_block_dest = perm_block_dest;
     p->num_perm_blocks = num_perm_blocks;
     p->perm_block_size = perm_block_size;

     p->move_size = move_size;
     p->move = move;

     return p;
}

void transpose_shmem_destroy_plan(transpose_shmem_plan p)
{
     if (p) {
	  if (p->block_dest)
	       free(p->block_dest);
	  if (p->perm_block_dest)
	       free(p->perm_block_dest);
	  if (p->move)
	    free(p->move);
	  free(p);
     }
}

static void exchange_doubles(double *buf1, double *buf2, int n)
{
     int i;
     double t0,t1,t2,t3;

     for (i = 0; i < (n & 3); ++i) {
	  t0 = buf1[i];
	  buf1[i] = buf2[i];
	  buf2[i] = t0;
     }
     for (; i < n; i += 4) {
	  t0 = buf1[i];
	  t1 = buf1[i+1];
	  t2 = buf1[i+2];
	  t3 = buf1[i+3];
	  buf1[i] = buf2[i];
	  buf1[i+1] = buf2[i+1];
	  buf1[i+2] = buf2[i+2];
	  buf1[i+3] = buf2[i+3];
	  buf2[i] = t0;
	  buf2[i+1] = t1;
	  buf2[i+2] = t2;
	  buf2[i+3] = t3;
     }
}

static void do_permutation(double *data,
			   int *perm_block_dest,
			   int num_perm_blocks,
			   int perm_block_size)
{
     int start_block;

     for (start_block = 0; start_block < num_perm_blocks; ++start_block) {
	  int cur_block = start_block;
	  int new_block = perm_block_dest[start_block];
	  
	  while (new_block > 0 && new_block < num_perm_blocks &&
		 new_block != start_block) {
	       exchange_doubles(data + perm_block_size*start_block,
				data + perm_block_size*new_block,
				perm_block_size);
	       perm_block_dest[cur_block] = -1 - new_block;
	       cur_block = new_block;
	       new_block = perm_block_dest[cur_block];
	  }

	  perm_block_dest[cur_block] = -1 - new_block;
     }

     /* reset the permutation array: */
     for (start_block = 0; start_block < num_perm_blocks; ++start_block)
	  perm_block_dest[start_block] = -1 - perm_block_dest[start_block];
}

void copy_transpose_2d_arbitrary(double *in, double *out,
				 long nx, long ny, int el_size)
  /* in and out are pointers to 2d, nx by ny, C-ordered arrays
     of elements that are stored completely locally.  The transpose\
     of the matrix "in" is written to "out". Each element of the array
     actually consists of el_size double numbers.*/
{
  double *cur,*end;
  int cur_x = 0;
  long cur_step;
  
  cur_step = nx * el_size;
  
  cur = out;
  end = out + el_size*nx*ny;
  while (cur_x < nx*el_size) {
    memcpy(cur,in,el_size * sizeof(double));
    in += el_size;
    
    if ((cur += cur_step) >= end)
      cur = out + (cur_x += el_size);
  }
}

void transpose_shmem(transpose_shmem_plan p, 
		     double *local_data, double *work,
		     int el_size)
{
     int my_pe, n_pes;
     int nx, ny, local_nx, local_ny;
     transpose_shmem_block_dest *block_dest;
     int block_size, num_blocks, block;
     int *perm_block_dest, num_perm_blocks, perm_block_size;

     my_pe = p->my_pe;
     n_pes = p->n_pes;
     nx = p->nx;
     ny = p->ny;
     local_nx = p->local_nx;
     local_ny = p->local_ny;

     block_dest = p->block_dest;
     block_size = p->block_size;
     num_blocks = p->num_blocks;
     perm_block_dest = p->perm_block_dest;
     num_perm_blocks = p->num_perm_blocks;
     perm_block_size = p->perm_block_size;

     /* first, transpose the local data: */

     if (local_nx > 0)
          copy_transpose_2d_arbitrary(local_data,work,local_nx,ny,el_size);

     /* The following two lines are only necessary
	on the Cray T3D, and can (should?) be removed
	for the Cray T3E: */
     shmem_set_cache_inv();
     barrier();

     for (block = 0; block < num_blocks; ++block) 
       if (block_dest[block].send_size > 0) {
	 if (block_dest[block].dest_pe == my_pe)
	   memcpy(local_data + el_size * block_dest[block].dest_offset,
		  work + 
		  block_dest[block].block_num * block_size * el_size,
		  block_dest[block].send_size * el_size * sizeof(double));
	 else
	   shmem_put(local_data + 
		     el_size * block_dest[block].dest_offset,
		     work + 
		     block_dest[block].block_num * block_size * el_size,
		     block_dest[block].send_size * el_size,
		     block_dest[block].dest_pe);
       }
     
     barrier(); /* wait for completion */

     shmem_clear_cache_inv(); /* a nop except on the T3D */

     if (local_ny > 0)
	  do_permutation(local_data,perm_block_dest,num_perm_blocks,
			 perm_block_size*el_size);
}

